using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using Leap;
using Leap.Unity;
using GPUTools.Physics.Scripts.Behaviours;
using Battlehub.RTSaveLoad.UnityEngineNS.AINS;

namespace Stopper {
	public class Disembody : MVRScript {

        DAZMergedSkinV2 skin;
        DAZCharacterSelector characterSelector;

        BodyPartConfig hands = new BodyPartConfig() {
            materials = {14, 16 },
            maleMaterials = {14, 17},
            bones = {"rHand", "lHand"},
            jointsToDisable = {"rCollar", "lCollar"},
            controllers = {"rShoulderControl", "lShoulderControl"}
        };

        BodyPartConfig arms = new BodyPartConfig() {
            materials = {14, 16, 22},
            maleMaterials = {14, 17, 23},
            bones = {"rForeArm", "lForeArm"},
            jointsToDisable = {"rCollar", "lCollar"},
            controllers = {"rShoulderControl", "lShoulderControl"}
        };

        BodyPartConfig feet = new BodyPartConfig() {
            materials = {12, 23},
            maleMaterials = {12, 24},
            bones = {"rFoot", "lFoot"},
            jointsToDisable = {"rThigh", "lThigh"},
            controllers = {"rThighControl", "lThighControl"}
        };


        BodyPartConfig penis = new BodyPartConfig() {
            materials = {},
            maleMaterials = {29},
            bones = {"hip", "Gen1", "Testes"},
            jointsToDisable = {"abdomen","rThigh", "lThigh"},
            controllers = {"abdomenControl", "rThighControl", "lThighControl"}
        };

        BodyPartConfig head = new BodyPartConfig() {
            materials = {1,2,3,4,5,6,7,8,9,10,11,13,15,24,25,26,27},
            maleMaterials = {1,2,3,4,5,6,7,8,9,10,11,13,15,16,25,26,27},
            bones = {"neck"},
            jointsToDisable = {"chest"},
            controllers = {"chestControl"}
        };

        List<BodyPartConfig> bodyparts;

        HashSet<GameObject> enabledBodyParts;
        bool[] enabledMaterials;

        Dictionary<Rigidbody, RigidBodyData> disabledRigidbodies = new Dictionary<Rigidbody, RigidBodyData>();
        Dictionary<ConfigurableJoint, JointData> disabledJoints = new Dictionary<ConfigurableJoint, JointData>();
        Dictionary<FreeControllerV3, ControllerData> disabledControllers = new Dictionary<FreeControllerV3, ControllerData>();
        
        HashSet<Collider> disabledColliders = new HashSet<Collider>();
        HashSet<AutoCollider> disabledAutoColliders = new HashSet<AutoCollider>();
        HashSet<MonoBehaviour> disabledGpuColliders = new HashSet<MonoBehaviour>();

        JSONStorableBool adjustMassAndDrag;

        public class JointData {
            public ConfigurableJointMotion xMotion;
            public ConfigurableJointMotion yMotion;
            public ConfigurableJointMotion zMotion;

            public ConfigurableJointMotion angularXMotion;
            public ConfigurableJointMotion angularYMotion;
            public ConfigurableJointMotion angularZMotion;

            public JointDrive slerpDrive;
        }

        public class RigidBodyData {
            public float mass;
            public float drag;
        }
        
        public class ControllerData {
            public FreeControllerV3.PositionState currentPositionState;
            public FreeControllerV3.RotationState currentRotationState;
            public float jointRotationDriveSpring;
            public float jointRotationDriveMaxForce;
        }


        public class BodyPartConfig {
            public bool enabled;
            public List<int> materials = new List<int>();
            public List<int> maleMaterials = new List<int>();
            public List<string> bones = new List<string>();
            public List<string> jointsToDisable = new List<string>();
            public List<string> controllers = new List<string>();
        }

        public override void Init() {
            bodyparts = new List<BodyPartConfig>{head, hands, arms, feet, penis};

            CreateToggle("Head", head, false);

            CreateToggle("Hands", hands, true);

            CreateToggle("Arms", arms, false);

            CreateToggle("Feet", feet, true);

            CreateToggle("Penis", penis, false);



            adjustMassAndDrag = new JSONStorableBool("Adjust Mass And Drag", false);
            RegisterBool(adjustMassAndDrag);
            CreateToggle(adjustMassAndDrag, false);
            
            AddSlider("Hidden Body Mass", 0.1f, 0.001f, 1.0f, SetHiddenBodyMass, true, "F3");
            AddSlider("Hidden Body Drag", 0f, 0f, 10f, SetHiddenBodyDrag, true, "F2");
            
			StartCoroutine("Initialize");
		}

        private void CreateToggle(string label, BodyPartConfig bodyPartConfig, bool defaultVal = false) {
            bodyPartConfig.enabled = defaultVal;

            var storableBool = new JSONStorableBool(label, defaultVal);
            RegisterBool(storableBool);
            var toggle = CreateToggle(storableBool, false);
            toggle.toggle.onValueChanged.AddListener((val) => {
                bodyPartConfig.enabled = val;
                OnDisable();
                OnEnable();
            });
        }

        private JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, string valueFormat = "F2", bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
            uiSlider.valueFormat = valueFormat;
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }
            
            return storableFloat;
        }


        private IEnumerator Initialize() {
            characterSelector = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
            if(characterSelector == null) {
                SuperController.LogError($"Could not find character selector. Is this a Person atom?");
                yield break;
            }

            while(skin == null) {
                skin = characterSelector.selectedCharacter.GetComponentInChildren<DAZMergedSkinV2>();
                yield return null;
            }
            
            SyncSettings();
        }

        private void SyncSettings() {
            if(skin == null) {
                return;
            }

            characterSelector = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
            if(characterSelector == null) {
                SuperController.LogError($"Could not find character selector. Is this a Person atom?");
                return;
            }

            var physicsModel = containingAtom.GetStorableByID("PhysicsModel");

            // Disable all skin materials
            for(int i = 0; i < skin.materialsEnabled.Length; i++) {
            //    SuperController.LogMessage($"{i} {skin.GPUmaterials[i].name}");
                skin.materialsEnabled[i] = false;
            }
            
            enabledBodyParts = new HashSet<GameObject>();
            enabledMaterials = new bool[skin.materialsEnabled.Length];

            foreach(var bodypart in bodyparts) {
                if(bodypart.enabled) {
                    foreach(string boneName in bodypart.bones) {
                        var bone = containingAtom.GetStorableByID(boneName);
                        enabledBodyParts.Add(bone.gameObject);
                    }
                    
                    List<int> materials = characterSelector.selectedCharacter.isMale ? bodypart.maleMaterials : bodypart.materials;
                    
                    foreach(int materialIndex in materials) {
                        skin.materialsEnabled[materialIndex] = true;
                        enabledMaterials[materialIndex] = true;
                    }
                    
                    foreach(var fc in containingAtom.freeControllers) {
                        if(bodypart.controllers.Contains(fc.name)) {
                            ToggleFreeController(fc, false);
                        }
                    }  
                    foreach(string boneName in bodypart.jointsToDisable) {
                        var bone = containingAtom.GetStorableByID(boneName);
                        foreach(ConfigurableJoint joint in bone.transform.GetComponents<ConfigurableJoint>()) {
                            ToggleJoint(joint, false);
                        }
                    }
                }
            }

            var hip = containingAtom.GetStorableByID("hip");
            DisablePhysicsRecursively(hip.gameObject);

            var breastSoftPhysics = containingAtom.GetStorableByID("BreastPhysicsMesh") as DAZPhysicsMesh;
            if(breastSoftPhysics != null)
			    breastSoftPhysics.gameObject.SetActive(false);

			var lowerBodySoftPhysics = containingAtom.GetStorableByID("LowerPhysicsMesh") as DAZPhysicsMesh;
            if(lowerBodySoftPhysics != null)
			    lowerBodySoftPhysics.gameObject.SetActive(false);
        }

        private void DisablePhysicsRecursively(GameObject go) {
            if(enabledBodyParts.Contains(go) && go.name != "hip") {
                return;
            }

            foreach(Rigidbody rb in go.GetComponents<Rigidbody>()) {
                ToggleRigidbody(rb, false);
            }

            foreach(Collider c in go.GetComponents<Collider>()) {
                ToggleCollider(c, false);
            }

            foreach(AutoCollider ac in go.GetComponents<AutoCollider>()) {
                ToggleAutoCollider(ac, false);
            }

            foreach(CapsuleLineSphereCollider c in go.GetComponents<CapsuleLineSphereCollider>()) {
                ToggleGpuCollider(c, false);
            }

            foreach(GpuSphereCollider c in go.GetComponents<GpuSphereCollider>()) {
                ToggleGpuCollider(c, false);
            }

            foreach(var child in go.transform.GetChildren()) {
                DisablePhysicsRecursively(child.gameObject);
            }
        }
		
        private void ToggleFreeController(FreeControllerV3 fc, bool on) {
            if(on && disabledControllers.ContainsKey(fc)) {
                ControllerData data = disabledControllers[fc];

                fc.jointRotationDriveSpring = data.jointRotationDriveSpring;
                fc.jointRotationDriveMaxForce = data.jointRotationDriveMaxForce;

                disabledControllers.Remove(fc);
            } else if (!on) {
                if(!disabledControllers.ContainsKey(fc))
                    disabledControllers[fc] = BuildControllerData(fc);

                fc.jointRotationDriveSpring = 0f;
                fc.jointRotationDriveMaxForce = 0f;
            }
        }

        private void ToggleCollider(Collider collider, bool on) {
            if(on) {
                collider.enabled = true;
                disabledColliders.Remove(collider);
            } else if (collider.enabled) {
                collider.enabled = false;
                disabledColliders.Add(collider);
            }
        }

        private void ToggleAutoCollider(AutoCollider ac, bool on) {
            if(on) {
                //ac.enabled = true;
                //ac.allowBatchUpdate = true;
                ac.gameObject.SetActive(true);
                disabledAutoColliders.Remove(ac);
            } else {
                ac.gameObject.SetActive(false);
                //ac.enabled = false;
                //ac.allowBatchUpdate = false;
                disabledAutoColliders.Add(ac);
            }
        }


        private void ToggleRigidbody(Rigidbody rb, bool on) {
            if(on && disabledRigidbodies.ContainsKey(rb)) {
                rb.detectCollisions = true;
                
                RigidBodyData rbData = disabledRigidbodies[rb];
                rb.mass = rbData.mass;
                rb.drag = rbData.drag;
                disabledRigidbodies.Remove(rb);
            } else if (rb.detectCollisions) {
                if(!disabledRigidbodies.ContainsKey(rb))
                    disabledRigidbodies[rb] = new RigidBodyData{mass = rb.mass, drag = rb.drag};

                rb.detectCollisions = false;

                if(adjustMassAndDrag.val) {
                    rb.mass = 0.1f;
                    rb.drag = 0f;
                }
            }
        }

        private void ToggleJoint(ConfigurableJoint joint, bool on) {
            if(on && disabledJoints.ContainsKey(joint)) {
                JointData jointData = disabledJoints[joint];
                
                joint.xMotion = jointData.xMotion;
                joint.yMotion = jointData.yMotion;
                joint.zMotion = jointData.zMotion;

                joint.angularXMotion = jointData.angularXMotion;
                joint.angularYMotion = jointData.angularYMotion;
                joint.angularZMotion = jointData.angularZMotion;

                joint.slerpDrive = jointData.slerpDrive;

                disabledJoints.Remove(joint);
            } else if (!on) {
                if(!disabledJoints.ContainsKey(joint))
                    disabledJoints[joint] = BuildJointData(joint);

                joint.xMotion = ConfigurableJointMotion.Free;
                joint.yMotion = ConfigurableJointMotion.Free;
                joint.zMotion = ConfigurableJointMotion.Free;

                joint.angularXMotion = ConfigurableJointMotion.Free;
                joint.angularYMotion = ConfigurableJointMotion.Free;
                joint.angularZMotion = ConfigurableJointMotion.Free;

                joint.rotationDriveMode = RotationDriveMode.Slerp;
                var slerpDrive = new JointDrive();
                slerpDrive.maximumForce = 0;
                slerpDrive.positionSpring = 0;
                joint.slerpDrive = slerpDrive;
            }
        }

        private void ToggleGpuCollider(MonoBehaviour collider, bool on) {
            if(on) {
                collider.enabled = true;
                disabledGpuColliders.Remove(collider);
            } else if (collider.enabled) {
                collider.enabled = false;
                disabledGpuColliders.Add(collider);
            }
        }

        private ControllerData BuildControllerData(FreeControllerV3 fc) {
            ControllerData data = new ControllerData();

            data.currentPositionState = fc.currentPositionState;
            data.currentRotationState = fc.currentRotationState;
            data.jointRotationDriveSpring = fc.jointRotationDriveSpring;
            data.jointRotationDriveMaxForce = fc.jointRotationDriveMaxForce;
            return data;
        }
        
        private JointData BuildJointData(ConfigurableJoint joint) {
            JointData jointData = new JointData();
            jointData.xMotion = joint.xMotion;
            jointData.yMotion = joint.yMotion;
            jointData.zMotion = joint.zMotion;

            jointData.angularXMotion = joint.angularXMotion;
            jointData.angularYMotion = joint.angularYMotion;
            jointData.angularZMotion = joint.angularZMotion;

            jointData.slerpDrive = joint.slerpDrive;

            return jointData; 
        }

        private void SetHiddenBodyMass(float mass) {
            if(adjustMassAndDrag.val) {
                foreach(var entry in disabledRigidbodies) {
                    entry.Key.mass = mass;
                }
            }
        }

        private void SetHiddenBodyDrag(float drag) {
            if(adjustMassAndDrag.val) {
                foreach(var entry in disabledRigidbodies) {
                    entry.Key.drag = drag;
                }
            }
        }

        private void Update() {
            if(skin != null && enabledMaterials != null) {
                for(int i = 0; i < skin.materialsEnabled.Length; i++){
                    skin.materialsEnabled[i] = enabledMaterials[i];
                }
            }
        }

        private void OnEnable() {
            SyncSettings();
        }
		
		private void OnDisable()
		{
            if(skin != null) {
                for(int i = 0; i < skin.materialsEnabled.Length; i++) {
                    skin.materialsEnabled[i] = true;
                }
            }
            
            // Disable all controllers
            foreach(var fc in disabledControllers.Keys.ToList()) {
                ToggleFreeController(fc, true);
            }

            foreach(Collider c in disabledColliders.ToList()) {
                ToggleCollider(c, true);
            }

            foreach(AutoCollider ac in disabledAutoColliders.ToList()) {
                ToggleAutoCollider(ac, true);
            }

            foreach(MonoBehaviour c in disabledGpuColliders.ToList()) {
                ToggleGpuCollider(c, true);
            }

            foreach(var joint in disabledJoints.Keys.ToList()) {
                ToggleJoint(joint, true);
            }

            foreach(var rb in disabledRigidbodies.Keys.ToList()) {
                ToggleRigidbody(rb, true);
            }

            var breastSoftPhysics = containingAtom.GetStorableByID("BreastPhysicsMesh") as DAZPhysicsMesh;
            if(breastSoftPhysics != null)
			    breastSoftPhysics.gameObject.SetActive(true);

			var lowerBodySoftPhysics = containingAtom.GetStorableByID("LowerPhysicsMesh") as DAZPhysicsMesh;
            if(lowerBodySoftPhysics != null)
			    lowerBodySoftPhysics.gameObject.SetActive(true);
		}
	}
}